#define FASTLED_ESP8266_RAW_PIN_ORDER
#define pinTemp D1 // Pin DHT-11
#define LedMult 1 // Moltiplicatore Luce
#define lcdBkl 3 // Pin Retro illuminazione
#define NUM_LEDS 12 // Numero di LED
#define DATA_PIN D4 // Pin Dati LED
#define bkTimeON 7 // Ora di Accensione della retroilluminazione
#define bkTimeOFF 22 // Ora di spegniemnto della retroilluminazione
#define TIME_OFFSET 1UL * 3600UL // UTC + 1 hour
#define MAX_HOURS 1
#define MAX_DAYS 1

// Librerie da includere
#include <Arduino.h>
#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
#include <WebSocketsClient.h> //  https://github.com/kakopappa/sinric/wiki/How-to-add-dependency-libraries
#include <ArduinoJson.h> // https://github.com/kakopappa/sinric/wiki/How-to-add-dependency-libraries (use the correct version)
#include <StreamString.h>
#include <DHTesp.h>
#include <NTPClient.h>
#include <WiFiUdp.h>
#include "FastLED.h"
#include <LiquidCrystal.h>
#include <TimeLib.h>
#include <time.h>
#include <Timezone.h>
#include <ArduinoOTA.h>
#include "Adafruit_MQTT.h"
#include "Adafruit_MQTT_Client.h"
//#include <ESP8266TelegramBOT.h>
#include <JSON_Decoder.h>
#include <OpenWeather.h>
#include <stdio.h>
#include <math.h>

#define PI  3.1415926535897932384626433832795
#define RAD (PI/180.0)
#define SMALL_FLOAT (1e-12)

byte customCharHum[] = {
  B00000,
  B00110,
  B01100,
  B10100,
  B10010,
  B10001,
  B10001,
  B01110
};
byte customCharTemp[] = {
  B00100,
  B01010,
  B01010,
  B01010,
  B01010,
  B10001,
  B10001,
  B01110
};

// Caratteri Luna
byte customCharLuna0[] = {
  B00000,
  B00100,
  B00000,
  B00000,
  B00000,
  B00000,
  B00100,
  B00000
};
byte customCharLuna1[] = {
  B00000,
  B00100,
  B00010,
  B00001,
  B00001,
  B00010,
  B00100,
  B00000
};
byte customCharLuna2[] = {
  B00000,
  B00100,
  B00110,
  B00111,
  B00111,
  B00110,
  B00100,
  B00000
};
byte customCharLuna3[] = {
  B00000,
  B00100,
  B01110,
  B01111,
  B01111,
  B01110,
  B00100,
  B00000
};
byte customCharLuna4[] = {
  B00000,
  B00100,
  B01110,
  B11111,
  B11111,
  B01110,
  B00100,
  B00000
};
byte customCharLuna5[] = {
  B00000,
  B00100,
  B01110,
  B11110,
  B11110,
  B01110,
  B00100,
  B00000
};
byte customCharLuna6[] = {
  B00000,
  B00100,
  B01100,
  B11100,
  B11100,
  B01100,
  B00100,
  B00000
};
byte customCharLuna7[] = {
  B00000,
  B00100,
  B01000,
  B10000,
  B10000,
  B01000,
  B00100,
  B00000
};

//Caratteri Meteo
byte customCharVuoto[] = {
  B00000,
  B00000,
  B00000,
  B00000,
  B00000,
  B00000,
  B00000,
  B00000
};
byte customCharPioggia[] = {
  B00000,
  B01110,
  B11111,
  B11111,
  B01110,
  B00000,
  B01010,
  B10100
};
byte customCharNuvolo[] = {
  B00000,
  B01110,
  B11111,
  B11111,
  B01110,
  B00000,
  B00000,
  B00000
};
byte customCharSole[] = {
  B00000,
  B00000,
  B10101,
  B01110,
  B11111,
  B01110,
  B10101,
  B00000
};
byte customCharNeve[] = {
  B00000,
  B00000,
  B10101,
  B00100,
  B11011,
  B00100,
  B10101,
  B00000
};
byte customCharSoleNuvola[] = {
  B00000,
  B00100,
  B10101,
  B01110,
  B11111,
  B11111,
  B01110,
  B00000
};
byte customCharNebbia[] = {
  B00000,
  B00000,
  B01010,
  B10101,
  B00000,
  B01010,
  B10101,
  B00000
};
byte customCharTemporale[] = {
  B01110,
  B11111,
  B11111,
  B01110,
  B01001,
  B10010,
  B01001,
  B10010
};

// ------------- Parametri Telegram --------------------------------------------------------------------
//#define BOTtoken ""  //token of FlashledBOT
//#define BOTname ""
//#define BOTusername ""
//int Bot_mtbs = 1000; //mean time between scan messages

// ------------ Parametri LCD --------------------------------------------------------------------------
const int RS = D2, EN = D0, d4 = D5, d5 = D6, d6 = D7, d7 = D3;

// ----------- Parametri WiFi --------------------------------------------------------------------------
#define MySSID "" // Change to your Wifi network SSID
#define MyWifiPassword "" // Change to your Wifi network password

//------------ Parametri SINRIC ------------------------------------------------------------------------
#define MyApiKey "" // Change to your sinric API Key. Your API Key is displayed on sinric.com dashboard
#define DeviceID "" // ID Controller RGB
#define DeviceID2 "" // ID Sensore Temperatura
#define SERVER_URL "iot.sinric.com" //"iot.sinric.com"
#define SERVER_PORT 80 // 80

// ----------- Parametri meteo -------------------------------------------------------------------------
//open weather map api key
String api_key = "";
//the city you want the weather for
String location = "mantova,IT";
int status = WL_IDLE_STATUS;
char owm_server[] = "api.openweathermap.org";
// Set both your longitude and latitude to at least 4 decimal places
String latitude =  "45.1721"; // 90.0000 to -90.0000 negative for Southern hemisphere
String longitude = "10.7856"; // 180.000 to -180.000 negative for West
String units = "metric";  // or "imperial"
String language = "en";   // See notes tab

// --------- Frequenza aggiornamenti --------------------------------------------------------------------
#define HEARTBEAT_INTERVAL 3*60*1000 // 3 Minutes 
#define TEMPRATURE_INTERVAL 1*60*1000 // 1 Minutes
#define IOT_INTERVAL 5*60*1000 // 5 Minutes

// --------- Information to Connect to Adafruit IO ------------------------------------------------------
#define AIO_SERVER      "io.adafruit.com"
#define AIO_SERVERPORT  1883                   // use 8883 for SSL
#define AIO_USERNAME    ""
#define AIO_KEY         ""
#define TEMP_FEED       "/feeds/temperatura"
#define HUM_FEED        "/feeds/umidita"

// ------------- Parametri Telegram --------------------------------------------------------------------
//TelegramBOT bot(BOTtoken, BOTname, BOTusername);
//long Bot_lasttime;   //last time messages' scan has been done
//bool Start = false;

// ------------- Parametri Display ---------------------------------------------------------------------
LiquidCrystal lcd(RS, EN, d4, d5, d6, d7);

//-------------- Parametri ora solare/legale -----------------------------------------------------------
static tm getDateTimeByParams(long time) {
  struct tm *newtime;
  const time_t tim = time;
  newtime = localtime(&tim);
  return *newtime;
}
static String getDateTimeStringByParams(tm *newtime, char* pattern = (char *)"%d/%m/%Y %H:%M:%S") {
  char buffer[30];
  strftime(buffer, 30, pattern, newtime);
  return buffer;
}
static String getEpochStringByParams(long time, char* pattern = (char *)"%d/%m/%Y %H:%M:%S") {
  //    struct tm *newtime;
  tm newtime;
  newtime = getDateTimeByParams(time);
  return getDateTimeStringByParams(&newtime, pattern);
}

// ------------ Inizializza WiFi -----------------------------------------------------------------------
ESP8266WiFiMulti WiFiMulti;
WebSocketsClient webSocket;
WiFiClient client;

//------------ Parametri Orologio ----------------------------------------------------------------------
const long utcOffsetInSeconds = 3600;
char daysOfTheWeek[7][12] = {"DOM", "LUN", "MAR", "MER", "GIO", "VEN", "SAB"};
// Define NTP Client to get time
WiFiUDP ntpUDP;
//NTPClient timeClient(ntpUDP, "192.168.1.254", utcOffsetInSeconds);
int GTMOffset = 0;
NTPClient timeClient(ntpUDP, "192.168.1.254", GTMOffset * 60 * 60, 60 * 60 * 1000);

// Central European Time (Frankfurt, Paris)
TimeChangeRule CEST = {"CEST", Last, Sun, Mar, 2, 120};     // Central European Summer Time
TimeChangeRule CET = {"CET ", Last, Sun, Oct, 3, 60};       // Central European Standard Time
Timezone CE(CEST, CET);

// --------- Stringhe IOT -------------------------------------------------------------------------------
Adafruit_MQTT_Client mqtt(&client, AIO_SERVER, AIO_SERVERPORT, AIO_USERNAME, AIO_KEY);
Adafruit_MQTT_Publish FeedTemperatura = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME TEMP_FEED);
Adafruit_MQTT_Publish FeedUmidita = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME HUM_FEED);

// --------- Non toccare da qui in poi ------------------------------------------------------------------
uint64_t heartbeatTimestamp = 0;
uint64_t tempratureUpdateTimestamp = 0;
uint64_t bkTimer = 0;
uint64_t TempUpdateTimestamp = 0;
uint64_t iotUpdateTimestamp = 0;
uint64_t TimeTimestamp = 0;
uint64_t StripTimestamp = 0;
bool isConnected = false;
bool luce = false;
bool oldState = false;
bool bkTemp = false;
bool arcobaleno = false;

void setPowerStateOnServer(String deviceId, String value);
void setSetTemperatureSettingOnServer(String deviceId, String value, String scale);
void readTempature();

float humidity;
float temperature;
float temperaturefh;

float hue = 0.0;
float sat = 0.0;
float bri = 0.0;
int brigh = 100;

OW_Weather ow; // Weather forecast library instance

DHTesp dht;
CRGB leds[NUM_LEDS];

void MQTT_connect() {
  int8_t ret;

  // Stop if already connected.
  if (mqtt.connected()) {
    return;
  } else {
    lcd.setCursor(15, 1);
    lcd.print("M");
  }

  Serial.print("Connecting to MQTT... ");

  uint8_t retries = 3;
  while ((ret = mqtt.connect()) != 0) { // connect will return 0 for connected
    Serial.println(mqtt.connectErrorString(ret));
    Serial.println("Retrying MQTT connection in 5 seconds...");
    mqtt.disconnect();
    delay(5000);  // wait 5 seconds
    retries--;
    if (retries == 0) {
      // basically die and wait for WDT to reset me
      //while (1);
      lcd.setCursor(15, 1);
      lcd.print("M");
      return;
    }
  }
  Serial.println("MQTT Connected!");
  lcd.setCursor(0, 0);
  lcd.print(" ");
}

void setLedColorHSV(int h, double s, double v) {
  //this is the algorithm to convert from RGB to HSV
  double r = 0;
  double g = 0;
  double b = 0;

  double hf = h / 60.0;

  int i = (int)floor(h / 60.0);
  double f = h / 60.0 - i;
  double pv = v * (1 - s);
  double qv = v * (1 - s * f);
  double tv = v * (1 - s * (1 - f));

  switch (i)
  {
    case 0: //rojo dominante
      r = v;
      g = tv;
      b = pv;
      break;
    case 1: //verde
      r = qv;
      g = v;
      b = pv;
      break;
    case 2:
      r = pv;
      g = v;
      b = tv;
      break;
    case 3: //azul
      r = pv;
      g = qv;
      b = v;
      break;
    case 4:
      r = tv;
      g = pv;
      b = v;
      break;
    case 5: //rojo
      r = v;
      g = pv;
      b = qv;
      break;
  }

  //set each component to a integer value between 0 and 255
  int red = constrain((int)255 * r, 0, 255);
  int green = constrain((int)255 * g, 0, 255);
  int blue = constrain((int)255 * b, 0, 255);

  setLedColor(red, green, blue);
}

void setLedColor(int red, int green, int blue) {
  //Sets the current color for the RGB LED
  int valR = (((red * brigh) / 100) * LedMult);
  int valG = (((green * brigh) / 100) * LedMult);
  int valB = (((blue * brigh) / 100) * LedMult);
  for (int i = 0; i <= NUM_LEDS; i++) {
    leds[i] = CRGB(valR, valG, valB);
  }

  Serial.print("R:");
  Serial.print(valR);
  Serial.print(" G:");
  Serial.print(valG);
  Serial.print(" B:");
  Serial.println(valB);

  FastLED.show();
}

void readTempature() {
  //Lettura temperatura da DHT
  //delay(dht.getMinimumSamplingPeriod());

  humidity = dht.getHumidity();
  temperature = dht.getTemperature();
  temperaturefh = dht.toFahrenheit(temperature);

  /*
    Serial.print(dht.getStatusString());
    Serial.print("\t");
    Serial.print(humidity, 1);
    Serial.print("\t\t");
    Serial.print(temperature, 1);
    Serial.print("\t\t");
    Serial.print(dht.toFahrenheit(temperature), 1);
    Serial.print("\t\t");
    Serial.print(dht.computeHeatIndex(temperature, humidity, false), 1);
    Serial.print("\t\t");
    Serial.println(dht.computeHeatIndex(dht.toFahrenheit(temperature), humidity, true), 1);
  */
}

void setPowerStateOnServer(String deviceId, String value) {
  // If you are going to use a push button to on/off the switch manually, use this function to update the status on the server
  // so it will reflect on Alexa app.
  // eg: setPowerStateOnServer("deviceid", "ON")

  // Call ONLY If status changed. DO NOT CALL THIS IN loop() and overload the server.
  DynamicJsonDocument root(1024);
  root["deviceId"] = deviceId;
  root["action"] = "setPowerState";
  root["value"] = value;
  StreamString databuf;
  serializeJson(root, databuf);

  webSocket.sendTXT(databuf);
}

void setSetTemperatureSettingOnServer(String deviceId, float setPoint, String scale, float ambientTemperature, float ambientHumidity) {
  // Call ONLY If status changed. DO NOT CALL THIS IN loop() and overload the server.
  //eg: setSetTemperatureSettingOnServer("deviceid", 25.0, "CELSIUS" or "FAHRENHEIT", 23.0, 45.3)
  // setPoint: Indicates the target temperature to set on the termostat.
  DynamicJsonDocument root(1024);
  root["action"] = "SetTemperatureSetting";
  root["deviceId"] = deviceId;

  JsonObject valueObj = root.createNestedObject("value");
  JsonObject temperatureSetting = valueObj.createNestedObject("temperatureSetting");
  temperatureSetting["setPoint"] = setPoint;
  temperatureSetting["scale"] = scale;
  temperatureSetting["ambientTemperature"] = ambientTemperature;
  temperatureSetting["ambientHumidity"] = ambientHumidity;

  StreamString databuf;
  serializeJson(root, databuf);

  //Serial.println(databuf);
  webSocket.sendTXT(databuf);
}

void setThermostatModeOnServer(String deviceId, String thermostatMode) {
  // Call ONLY If status changed. DO NOT CALL THIS IN loop() and overload the server.
  DynamicJsonDocument root(1024);
  root["deviceId"] = deviceId;
  root["action"] = "SetThermostatMode";
  root["value"] = thermostatMode;
  StreamString databuf;
  serializeJson(root, databuf);

  webSocket.sendTXT(databuf);
}

void spegniTutto() {
  for (int i = 0; i <= NUM_LEDS; i++) {
    leds[i] = CRGB(0, 0, 0);
  }
  FastLED.show();
}

double Julian(int year, int month, double day) {
  int a, b, c, e;
  if (month < 3) {
    year--;
    month += 12;
  }
  if (year > 1582 || (year == 1582 && month > 10) ||
      (year == 1582 && month == 10 && day > 15)) {
    a = year / 100;
    b = 2 - a + a / 4;
  }
  c = 365.25 * year;
  e = 30.6001 * (month + 1);
  return b + c + e + day + 1720994.5;
}

double sun_position(double j) {
  double n, x, e, l, dl, v;
  double m2;
  int i;

  n = 360 / 365.2422 * j;
  i = n / 360;
  n = n - i * 360.0;
  x = n - 3.762863;
  if (x < 0) x += 360;
  x *= RAD;
  e = x;
  do {
    dl = e - .016718 * sin(e) - x;
    e = e - dl / (1 - .016718 * cos(e));
  } while (fabs(dl) >= SMALL_FLOAT);
  v = 360 / PI * atan(1.01686011182 * tan(e / 2));
  l = v + 282.596403;
  i = l / 360;
  l = l - i * 360.0;
  return l;
}

double moon_position(double j, double ls) {
  double ms, l, mm, n, ev, sms, z, x, lm, bm, ae, ec;
  double d;
  int i;

  /* ls = sun_position(j) */
  ms = 0.985647332099 * j - 3.762863;
  if (ms < 0) ms += 360.0;
  l = 13.176396 * j + 64.975464;
  i = l / 360;
  l = l - i * 360.0;
  if (l < 0) l += 360.0;
  mm = l - 0.1114041 * j - 349.383063;
  i = mm / 360;
  mm -= i * 360.0;
  n = 151.950429 - 0.0529539 * j;
  i = n / 360;
  n -= i * 360.0;
  ev = 1.2739 * sin((2 * (l - ls) - mm) * RAD);
  sms = sin(ms * RAD);
  ae = 0.1858 * sms;
  mm += ev - ae - 0.37 * sms;
  ec = 6.2886 * sin(mm * RAD);
  l += ev + ec - ae + 0.214 * sin(2 * mm * RAD);
  l = 0.6583 * sin(2 * (l - ls) * RAD) + l;
  return l;
}

int moon_phase() {
  String dataOraString = getEpochStringByParams(CE.toLocal(now()));
  int year = dataOraString.substring(0, 5).toInt();
  int month = dataOraString.substring(6, 8).toInt();
  int day = dataOraString.substring(9, 11).toInt();
  double hour = 12.00;
  double j = Julian(year, month, (double)day + hour / 24.0) - 2444238.5;
  double ls = sun_position(j);
  double lm = moon_position(j, ls);

  double t = lm - ls;
  if (t < 0) t += 360;

  //*ip = (int)((t + 22.5)/45) & 0x7;        // Moon state 0-7 for moonPhase[] index
  return map((((int)((t + 7.5)/15) + 23) % 24), 0, 23, 0, 7);  // Moon state 0-23 for icon bitmap

  //return 100.0 * ((1.0 - cos((lm - ls) * RAD)) / 2) + 0.5; // percent illuminated
}

void stampaLuna() {
  int faseLuna = moon_phase();
  lcd.createChar(2, customCharVuoto);
  switch (faseLuna) {
    case 0:
      lcd.createChar(2, customCharLuna0);
      break;
    case 1:
      lcd.createChar(2, customCharLuna1);
      break;
    case 2:
      lcd.createChar(2, customCharLuna2);
      break;
    case 3:
      lcd.createChar(2, customCharLuna3);
      break;
    case 4:
      lcd.createChar(2, customCharLuna4);
      break;
    case 5:
      lcd.createChar(2, customCharLuna5);
      break;
    case 6:
      lcd.createChar(2, customCharLuna6);
      break;
    case 7:
      lcd.createChar(2, customCharLuna7);
      break;
    default:
      lcd.setCursor(0,0);
      lcd.write("X");
      break;
  }
  lcd.setCursor(0, 0);
  lcd.write(2);
}

/* void Bot_ExecMessages() {
  for (int i = 1; i < bot.message[0][0].toInt() + 1; i++)      {
    bot.message[i][5]=bot.message[i][5].substring(1,bot.message[i][5].length());
    if (bot.message[i][5] == "\/valvoleon") {
      luce = true;
      arcobaleno = true;
      bot.sendMessage(bot.message[i][4], "Valvole Accese", "");
    }
    if (bot.message[i][5] == "\/valvoleoff") {
      luce = false;
      arcobaleno = false;
      bot.sendMessage(bot.message[i][4], "Valvole Spente", "");
    }
    if (bot.message[i][5] == "\/temp") {
      String msgTemp;
      msgTemp += "Temperaura: ";
      msgTemp += String(temperature,2);
      msgTemp += "°C";
      String msgHum;
      msgHum += "Umidita: ";
      msgHum += String(temperature,1);
      msgHum += "%";
      bot.sendMessage(bot.message[i][4], msgTemp, "");
      bot.sendMessage(bot.message[i][4], msgHum, "");
    }
    if (bot.message[i][5] == "\/start") {
      String wellcome = "Ciao, benvenuto nel BOT per le tue Valvole";
      String wellcome1 = "/temp : Per ricevere temperatura e umidità";
      String wellcome2 = "/valvoleon : Per accendere la luce";
      String wellcome3 = "/valvoleff : Per spegnere la luce";
      bot.sendMessage(bot.message[i][4], wellcome, "");
      bot.sendMessage(bot.message[i][4], wellcome1, "");
      bot.sendMessage(bot.message[i][4], wellcome2, "");
      bot.sendMessage(bot.message[i][4], wellcome3, "");
      Start = true;
    }
  }
  bot.message[0][0] = "";   // All messages have been replied - reset new messages
  } */

String stampaOra() {
  String dataOra = "";
  int tempTime = 0;
  //timeClient.update();

  if (timeClient.update()) {
    lcd.setCursor(15, 1);
    lcd.print(" ");
    unsigned long epoch = timeClient.getEpochTime();
    setTime(epoch);
  } else {
    Serial.print ( "NTP Update not WORK!!" );
    lcd.setCursor(15, 1);
    lcd.print("T");
  }

  //Serial.println("GetEpoch: " + getEpochStringByParams(CE.toLocal(now())));

  dataOra += (daysOfTheWeek[timeClient.getDay()]);
  dataOra += " ";

  String dataOraString = getEpochStringByParams(CE.toLocal(now()));

  tempTime = dataOraString.substring(11, 13).toInt();
  if (((tempTime > bkTimeON) && (tempTime < bkTimeOFF)) || bkTemp) {
    digitalWrite(lcdBkl, HIGH);
  } else {
    digitalWrite(lcdBkl, LOW);
  }
  if (tempTime < 10) {
    dataOra += "0";
  }
  dataOra += tempTime;
  dataOra += ":";
  tempTime = dataOraString.substring(14, 16).toInt();
  if (tempTime < 10) {
    dataOra += "0";
  }
  dataOra += tempTime;
  dataOra += ":";
  tempTime = dataOraString.substring(17, 19).toInt();
  if (tempTime < 10) {
    dataOra += "0";
  }
  dataOra += tempTime;

  //Serial.println("getTime: " + dataOra);

  lcd.setCursor(2, 0);
  lcd.print(dataOra);
  //Serial.println(dataOra);

  return dataOra;
}

void stampaMeteo(int id) {
  //Serial.print("Ricevuto: ");
  //Serial.println(id);
  lcd.createChar(3, customCharVuoto);
  if (id >= 200 && id < 300) {
     //Temporale
     lcd.createChar(3, customCharTemporale);
     lcd.setCursor(15, 0);
     lcd.write(3);
     Serial.println("Temporale");
  } else if ( id >= 300 && id < 400) {
    //Drizzle
    lcd.createChar(3, customCharSoleNuvola);
    lcd.setCursor(15, 0);
    lcd.write(3);
    Serial.println("Pioviggine");
  } else if (id >= 500 && id < 600) {
    //Pioggia
    lcd.createChar(3, customCharPioggia);
    lcd.setCursor(15, 0);
    lcd.write(3);
    Serial.println("Pioggia");
  } else if (id >= 600 && id < 700) {
    //Neve
    lcd.createChar(3, customCharNeve);
    lcd.setCursor(15, 0);
    lcd.write(3);
    Serial.println("Neve");
  } else if (id >= 700 && id < 800) {
    //Nebbia
    lcd.createChar(3, customCharNebbia);
    lcd.setCursor(15, 0);
    lcd.write(3);
    Serial.println("Nebbia");
  } else if (id == 800) {
    //Sole
    lcd.createChar(3, customCharSole);
    lcd.setCursor(15, 0);
    lcd.write(3);
    Serial.println("Sole");
  } else if (id >= 801 && id < 900) {
    //Nuvolo
    lcd.createChar(3, customCharNuvolo);
    lcd.setCursor(15, 0);
    lcd.write(3);
    Serial.println("Nuvolo");
  } else {
    //Sconosciuto
    lcd.setCursor(15, 0); 
    lcd.print("X");
    Serial.println("Sconosciuto!");
  }
}

void fadeall() {
  for (int i = 0; i < NUM_LEDS; i++) {
    leds[i].nscale8(255);
  }
}

// deviceId is the ID assgined to your smart-home-device in sinric.com dashboard. Copy it from dashboard and paste it here
void turnOn(String deviceId) {
  if (deviceId == DeviceID) // Device ID of first device
  {
    Serial.print("Turn on device id: ");
    Serial.println(deviceId);
    luce = true;
    arcobaleno = true;
  }
  else {
    Serial.print("Turn on for unknown device id: ");
    Serial.println(deviceId);
  }
}

void turnOff(String deviceId) {
  if (deviceId == DeviceID) // Device ID of first device
  {
    Serial.print("Turn off Device ID: ");
    Serial.println(deviceId);
    luce = false;
    arcobaleno = false;
  }
  else {
    Serial.print("Turn off for unknown device id: ");
    Serial.println(deviceId);
  }
}

void webSocketEvent(WStype_t type, uint8_t * payload, size_t length) {
  switch (type) {
    case WStype_DISCONNECTED:
      isConnected = false;
      Serial.printf("[WSc] Webservice disconnected from sinric.com!\n");
      break;
    case WStype_CONNECTED: {
        isConnected = true;
        Serial.printf("[WSc] Service connected to sinric.com at url: %s\n", payload);
        Serial.printf("Waiting for commands from sinric.com ...\n");
      }
      break;
    case WStype_TEXT: {
        Serial.printf("[WSc] get text: %s\n", payload);
        // Example payloads

        // For Switch or Light device types
        // {"deviceId": xxxx, "action": "setPowerState", value: "ON"} // https://developer.amazon.com/docs/device-apis/alexa-powercontroller.html

        // For Light device type
        // Look at the light example in github

        DynamicJsonDocument json(1024);
        deserializeJson(json, (char*)payload);
        String deviceId = json ["deviceId"];
        String action = json ["action"];
        if (deviceId == DeviceID)
        {
          bkTimer = millis();
          bkTemp = true;

          if (action == "setPowerState") { // Switch or Light
            String value = json ["value"];
            if (value == "ON") {
              turnOn(deviceId);
            } else {
              turnOff(deviceId);
            }
          }
          else if (action == "SetColor") {
            arcobaleno = false;
            // Alexa, set the device name to red
            // get text: {"deviceId":"xxxx","action":"SetColor","value":{"hue":0,"saturation":1,"brightness":1}}
            String Ahue = json ["value"]["hue"];
            String Asaturation = json ["value"]["saturation"];
            String Abrightness = json ["value"]["brightness"];

            hue = Ahue.toInt();
            sat = Asaturation.toDouble();
            bri = Abrightness.toDouble();

            Serial.println("[WSc] hue: " + Ahue);
            Serial.println("[WSc] saturation: " + Asaturation);
            Serial.println("[WSc] brightness: " + Abrightness);
            luce = true;
          }
          else if (action == "SetColorTemperature") {
            arcobaleno = false;
            String strJson = json ["value"];
            switch (strJson.toInt()) {
              case 2200:
                hue = 29.0;
                sat = 0.82;
                bri = 1;
                break;
              case 2700:
                hue = 29.0;
                sat = 0.66;
                bri = 1;
                break;
              case 4000:
                hue = 30.0;
                sat = 0.36;
                bri = 1;
                break;
              case 5500:
                hue = 23.0;
                sat = 0.12;
                bri = 1;
                break;
              case 7000:
                hue = 250;
                sat = 0.04;
                bri = 11;
                break;
              default:
                String Ahue = strJson;
                String Asaturation = "0.8";
                String Abrightness = "1";

                hue = Ahue.toDouble() / 28.0;
                sat = Asaturation.toDouble();
                bri = Abrightness.toDouble();
                break;
            }

            /*
              String Ahue = json ["value"];
              String Asaturation = "0.8";
              String Abrightness = "1";

              hue = Ahue.toDouble() / 28.0;
              sat = Asaturation.toDouble();
              bri = Abrightness.toDouble();
            */
            luce = true;
          }
          else if (action == "SetBrightness") {
            String Abrightness = json ["value"];;

            brigh = Abrightness.toInt();
            luce = true;
          }
          else if (action == "test") {
            Serial.println("[WSc] received test command from sinric.com");
          }
        } else if (deviceId == DeviceID2) {
          Serial.println("[WSc] Ricevuto dato per termostato");
          readTempature();
          setSetTemperatureSettingOnServer(DeviceID2, temperature, "CELSIUS", temperature, humidity);
        }
        else {
          Serial.println("[WSc] Nope!");
        }
      }
      break;
    case WStype_BIN:
      Serial.printf("[WSc] get binary length: %u\n", length);
      break;
  }
}

void printCurrentWeather() {
  // Create the structures that hold the retrieved weather
  OW_current *current = new OW_current;
  OW_hourly *hourly = new OW_hourly;
  OW_daily  *daily = new OW_daily;

  Serial.print("\nRequesting weather information from OpenWeather... ");

  ow.getForecast(current, hourly, daily, api_key, latitude, longitude, units, language);
  //ow.getForecast(hourly, api_key, latitude, longitude, units, language);

  Serial.println("Weather from Open Weather\n");

  // Position as reported by Open Weather
  Serial.print("Latitude            : "); Serial.println(ow.lat);
  Serial.print("Longitude           : "); Serial.println(ow.lon);
  // We can use the timezone to set the offset eventually...
  Serial.print("Timezone            : "); Serial.println(ow.timezone);
  Serial.println();

  /*
  if (current) {
    Serial.println("############### Current weather ###############\n");
    Serial.print("dt (time)        : "); Serial.print(strTime(current->dt));
    Serial.print("sunrise          : "); Serial.print(strTime(current->sunrise));
    Serial.print("sunset           : "); Serial.print(strTime(current->sunset));
    Serial.print("temp             : "); Serial.println(current->temp);
    Serial.print("feels_like       : "); Serial.println(current->feels_like);
    Serial.print("pressure         : "); Serial.println(current->pressure);
    Serial.print("humidity         : "); Serial.println(current->humidity);
    Serial.print("dew_point        : "); Serial.println(current->dew_point);
    Serial.print("uvi              : "); Serial.println(current->uvi);
    Serial.print("clouds           : "); Serial.println(current->clouds);
    Serial.print("visibility       : "); Serial.println(current->visibility);
    Serial.print("wind_speed       : "); Serial.println(current->wind_speed);
    Serial.print("wind_gust        : "); Serial.println(current->wind_gust);
    Serial.print("wind_deg         : "); Serial.println(current->wind_deg);
    Serial.print("rain             : "); Serial.println(current->rain);
    Serial.print("snow             : "); Serial.println(current->snow);
    Serial.println();
    Serial.print("id               : "); Serial.println(current->id);
    Serial.print("main             : "); Serial.println(current->main);
    Serial.print("description      : "); Serial.println(current->description);
    Serial.print("icon             : "); Serial.println(current->icon);

    Serial.println();
  }
  if (hourly) {
    Serial.println("############### Hourly weather  ###############\n");
    for (int i = 0; i < MAX_HOURS; i++)
    {
      Serial.print("Hourly summary  "); if (i < 10) Serial.print(" "); Serial.print(i);
      Serial.println();
      Serial.print("dt (time)        : "); Serial.print(strTime(hourly->dt[i]));
      Serial.print("temp             : "); Serial.println(hourly->temp[i]);
      Serial.print("feels_like       : "); Serial.println(hourly->feels_like[i]);
      Serial.print("pressure         : "); Serial.println(hourly->pressure[i]);
      Serial.print("humidity         : "); Serial.println(hourly->humidity[i]);
      Serial.print("dew_point        : "); Serial.println(hourly->dew_point[i]);
      Serial.print("clouds           : "); Serial.println(hourly->clouds[i]);
      Serial.print("wind_speed       : "); Serial.println(hourly->wind_speed[i]);
      Serial.print("wind_gust        : "); Serial.println(hourly->wind_gust[i]);
      Serial.print("wind_deg         : "); Serial.println(hourly->wind_deg[i]);
      Serial.print("rain             : "); Serial.println(hourly->rain[i]);
      Serial.print("snow             : "); Serial.println(hourly->snow[i]);
      Serial.println();
      Serial.print("id               : "); Serial.println(hourly->id[i]);
      Serial.print("main             : "); Serial.println(hourly->main[i]);
      Serial.print("description      : "); Serial.println(hourly->description[i]);
      Serial.print("icon             : "); Serial.println(hourly->icon[i]);
      Serial.print("pop              : "); Serial.println(hourly->pop[i]);

      Serial.println();
    }
  }
  if (daily) {
    Serial.println("###############  Daily weather  ###############\n");
    for (int i = 0; i < MAX_DAYS; i++)
    {
      Serial.print("Daily summary   "); if (i < 10) Serial.print(" "); Serial.print(i);
      Serial.println();
      Serial.print("dt (time)        : "); Serial.print(strTime(daily->dt[i]));
      Serial.print("sunrise          : "); Serial.print(strTime(daily->sunrise[i]));
      Serial.print("sunset           : "); Serial.print(strTime(daily->sunset[i]));

      Serial.print("temp.morn        : "); Serial.println(daily->temp_morn[i]);
      Serial.print("temp.day         : "); Serial.println(daily->temp_day[i]);
      Serial.print("temp.eve         : "); Serial.println(daily->temp_eve[i]);
      Serial.print("temp.night       : "); Serial.println(daily->temp_night[i]);
      Serial.print("temp.min         : "); Serial.println(daily->temp_min[i]);
      Serial.print("temp.max         : "); Serial.println(daily->temp_max[i]);

      Serial.print("feels_like.morn  : "); Serial.println(daily->feels_like_morn[i]);
      Serial.print("feels_like.day   : "); Serial.println(daily->feels_like_day[i]);
      Serial.print("feels_like.eve   : "); Serial.println(daily->feels_like_eve[i]);
      Serial.print("feels_like.night : "); Serial.println(daily->feels_like_night[i]);

      Serial.print("pressure         : "); Serial.println(daily->pressure[i]);
      Serial.print("humidity         : "); Serial.println(daily->humidity[i]);
      Serial.print("dew_point        : "); Serial.println(daily->dew_point[i]);
      Serial.print("uvi              : "); Serial.println(daily->uvi[i]);
      Serial.print("clouds           : "); Serial.println(daily->clouds[i]);
      Serial.print("visibility       : "); Serial.println(daily->visibility[i]);
      Serial.print("wind_speed       : "); Serial.println(daily->wind_speed[i]);
      Serial.print("wind_gust        : "); Serial.println(daily->wind_gust[i]);
      Serial.print("wind_deg         : "); Serial.println(daily->wind_deg[i]);
      Serial.print("rain             : "); Serial.println(daily->rain[i]);
      Serial.print("snow             : "); Serial.println(daily->snow[i]);
      Serial.println();
      Serial.print("id               : "); Serial.println(daily->id[i]);
      Serial.print("main             : "); Serial.println(daily->main[i]);
      Serial.print("description      : "); Serial.println(daily->description[i]);
      Serial.print("icon             : "); Serial.println(daily->icon[i]);
      Serial.print("pop              : "); Serial.println(daily->pop[i]);

      Serial.println();
    }
  }
 */

  //Serial.print("ID Meteo: ");
  //Serial.println(daily->id[0]);
  stampaMeteo((int) hourly->id[0]);

  // Delete to free up space and prevent fragmentation as strings change in length
  delete current;
  delete hourly;
  delete daily;
}

String strTime(time_t unixTime) {
  /***************************************************************************************
  **                          Convert unix time to a time string
  ***************************************************************************************/

  unixTime += TIME_OFFSET;
  return ctime(&unixTime);
}

void setup() {
  Serial.begin(115200);
  Serial.println("Accensione...");

  lcd.begin(16, 2);
  lcd.setCursor(0, 0);
  lcd.print("Accensione...");
  lcd.createChar(0, customCharTemp);
  lcd.createChar(1, customCharHum);

  pinMode(lcdBkl, OUTPUT);
  digitalWrite(lcdBkl, HIGH);

  FastLED.addLeds<WS2811, DATA_PIN, BRG>(leds, NUM_LEDS);
  LEDS.setBrightness(84);

  dht.setup(pinTemp, DHTesp::DHT11); // Connect DHT sensor to GPIO

  WiFiMulti.addAP(MySSID, MyWifiPassword);
  Serial.println();
  Serial.print("Connecting to Wifi: ");
  Serial.println(MySSID);
  lcd.setCursor(0, 0);
  lcd.print("Connessione...");
  lcd.setCursor(0, 1);

  // Waiting for Wifi connect
  while (WiFiMulti.run() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
    lcd.print(".");

  }
  if (WiFiMulti.run() == WL_CONNECTED) {
    Serial.println("");
    Serial.print("WiFi connected. ");
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Connesso!");
    lcd.setCursor(0, 1);
    lcd.print(WiFi.localIP());
  }
  delay(2000);

  spegniTutto();

  // server address, port and URL
  webSocket.begin(SERVER_URL, SERVER_PORT, "/");

  // event handler
  webSocket.onEvent(webSocketEvent);
  webSocket.setAuthorization("apikey", MyApiKey);

  // try again every 5000ms if connection has failed
  webSocket.setReconnectInterval(5000);
  // If you see 'class WebSocketsClient' has no member named 'setReconnectInterval' error update arduinoWebSockets
  ArduinoOTA.onStart([]() {
    Serial.println("Start");
  });
  ArduinoOTA.onEnd([]() {
    Serial.println("\nEnd");
  });
  ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
    Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
  });
  ArduinoOTA.onError([](ota_error_t error) {
    Serial.printf("Error[%u]: ", error);
    if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
    else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
    else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
    else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
    else if (error == OTA_END_ERROR) Serial.println("End Failed");
  });
  ArduinoOTA.begin();
  Serial.print("OTA Update Ready");
  Serial.print(" IP address: ");
  Serial.println(WiFi.localIP());

  lcd.clear();

  //bot.begin();
}

void loop() {
  webSocket.loop();
  ArduinoOTA.handle();

  uint64_t now = millis();

  /*if (now > Bot_lasttime + Bot_mtbs)  {
    bot.getUpdates(bot.message[0][1]);   // launch API GetUpdates up to xxx message
    Bot_ExecMessages();   // reply to message with Echo
    Bot_lasttime = millis();
    } */

  if (isConnected) {
    // Send heartbeat in order to avoid disconnections during ISP resetting IPs over night. Thanks @MacSass
    if ((now - heartbeatTimestamp) > HEARTBEAT_INTERVAL) {
      heartbeatTimestamp = now;
      if (luce) {
        setPowerStateOnServer(DeviceID, "ON");
      } else {
        setPowerStateOnServer(DeviceID, "OFF");
      }
      //getWeather();
      printCurrentWeather();
    }
    if ((now - StripTimestamp) > 1000) {
      if (luce) {
        if (arcobaleno) {
          for (int i = 0; i < NUM_LEDS; i++) {
            // Set the i'th led to red
            leds[i] = CHSV(hue++, 255, 255);
            // Show the leds
            FastLED.show();
            // now that we've shown the leds, reset the i'th led to black
            // leds[i] = CRGB::Black;
            fadeall();
            delay(10);
          }
        } else {
          if ((hue == 0.0) && (sat == 0.0) && (bri == 0.0)) {
            hue = 120.0;
            sat = 100.0;
            bri = 100.0;
          }
          StripTimestamp = now;
          setLedColorHSV(hue, sat, bri);
        }
      } else {
        spegniTutto();
      }
    }
    if ((now - tempratureUpdateTimestamp) > TEMPRATURE_INTERVAL) {
      tempratureUpdateTimestamp = now;
      setSetTemperatureSettingOnServer(DeviceID2, temperature, "CELSIUS", temperature, humidity);
    }
    if (luce != oldState) {
      if (luce) {
        setPowerStateOnServer(DeviceID, "ON");
      } else {
        setPowerStateOnServer(DeviceID, "OFF");
      }
      oldState = luce;
    }
  }
  if ((now - iotUpdateTimestamp) > IOT_INTERVAL) {
    MQTT_connect();
    delay(500);
    iotUpdateTimestamp = now;
    FeedTemperatura.publish(temperature);
    FeedUmidita.publish(humidity);
    Serial.println("[IOT] Pubblicato");
    mqtt.disconnect();
  }
  if ((now - bkTimer) > 15000) {
    bkTemp = false;
  }
  if ((now - TempUpdateTimestamp) > 15000) {
    TempUpdateTimestamp = now;
    readTempature();
    stampaLuna();
  }
  if ((now - TimeTimestamp) > 999) {
    TimeTimestamp = now;
    //Serial.println(stampaOra());
    // Stampa Orario
    //lcd.setCursor(2, 0);
    //lcd.print(stampaOra());
    stampaOra();
    // Stampa Temperatura
    lcd.setCursor(1, 1);
    lcd.write(0);
    //lcd.setCursor(1, 1);
    //lcd.print("T:");
    lcd.setCursor(3, 1);
    lcd.print(int(temperature), DEC);
    lcd.print(".");
    if (int(temperature * 100) - (int(temperature) * 100) < 10) lcd.print("0");
    lcd.print(int(temperature * 100) - (int(temperature) * 100));
    lcd.print("C ");
    lcd.setCursor(10, 1);
    lcd.write(1);
    //lcd.print("H:");
    lcd.setCursor(12, 1);
    lcd.print(int(humidity), DEC);
    lcd.print("%");
    //ctrlBkLight();
  }
}
